@@ -51,6 +51,13 @@ from . import voucher
 from . import currency
 from . import order_line
 
+COUTRIES_MAP = dict(
+    ES = "Spain",
+    PT = "Portugal"
+)
+""" Map that associated ISO alpha-2 country codes with
+their corresponding full text version to be used in Omni """
+
 class Order(bundle.Bundle):
 
     STATUS_S = dict(
@@ -118,7 +125,9 @@ class Order(bundle.Bundle):
 
     email = appier.field(
         index = "hashed",
-        safe = True
+        safe = True,
+        observations = """The email of the entity owner of the current
+        order, typically a personal email address of the buyer"""
     )
 
     gift_wrap = appier.field(
@@ -562,6 +571,7 @@ class Order(bundle.Bundle):
     def pay_s(
         self,
         payment_data = None,
+        payment_function = None,
         vouchers = True,
         strict = True,
         notify = False,
@@ -570,7 +580,11 @@ class Order(bundle.Bundle):
         payment_data = payment_data or dict()
         if ensure_waiting: self.ensure_waiting_s()
         self.verify_paid()
-        result = self._pay(payment_data, strict = strict)
+        result = self._pay(
+            payment_data,
+            payment_function = payment_function,
+            strict = strict
+        )
         confirmed = result == True
         self.save()
         if vouchers: self.use_vouchers_s()
@@ -581,6 +595,7 @@ class Order(bundle.Bundle):
     def end_pay_s(
         self,
         payment_data = None,
+        payment_function = None,
         strict = False,
         notify = False
     ):
@@ -594,7 +609,11 @@ class Order(bundle.Bundle):
         # (finish payment) operation return a valid value in case
         # a concrete operation has been executed or an invalid
         # value in case nothing has been done
-        result = self._end_pay(payment_data, strict = strict)
+        result = self._end_pay(
+            payment_data,
+            payment_function = payment_function,
+            strict = strict
+        )
 
         # marks the current order as completely paid and in case
         # the notify flag is set notifies the event handlers
@@ -608,13 +627,18 @@ class Order(bundle.Bundle):
     def cancel_s(
         self,
         cancel_data = None,
+        cancel_function = None,
         vouchers = True,
         strict = False,
         notify = False
     ):
         cancel_data = cancel_data or dict()
         cancel_data.update(self.cancel_data)
-        result = self._cancel(cancel_data, strict = strict)
+        result = self._cancel(
+            cancel_data,
+            cancel_function = cancel_function,
+            strict = strict
+        )
         self.mark_canceled_s()
         if vouchers: self.disuse_vouchers_s()
         if notify: self.notify_s()
@@ -702,7 +726,7 @@ class Order(bundle.Bundle):
         if self.paid and not force: return
 
         # in case the discount data is not valid there's no voucher
-        # value to be reverted/disused, should return immediately
+        # value to be reverted/discounted, should return immediately
         if not self.discount_data: return
 
         # iterates over the complete set of vouchers and associated amount
@@ -1013,6 +1037,174 @@ class Order(bundle.Bundle):
         if not self.is_closed(): return
         self.close_lines_s()
 
+    @appier.operation(
+        name = "Import Omni",
+        parameters = (
+            ("Invoice", "invoice", bool, False),
+            ("Strict", "strict", bool, True)
+        ),
+        level = 2
+    )
+    def import_omni_s(self, invoice = False, strict = True):
+        api = self.owner.get_omni_api()
+        appier.verify(
+            self.paid,
+            message = "Order is not yet paid"
+        )
+
+        # verifies if the current order is already "marked" with the
+        # Omni import timestamp, if that's the case and the strict mode
+        # is enabled then an operation error is raised
+        omni_timestamp = self.meta.get("omni_timestamp", None)
+        if strict and omni_timestamp: raise appier.OperationalError(
+            message = "Order already imported in Omni"
+        )
+
+        # retrieves via configuration the store from which the
+        # sale is going to be performed (affects inventory)
+        store_id = appier.conf("OMNI_BOT_STORE", None)
+
+        # builds the "unique" description of the order from which
+        # a duplicates are going to be avoided by explicit checking
+        description = "budy:order:%s" % self.reference
+        orders = api.list_sales(
+            number_records = 1,
+            **{
+                "filters[]" : [
+                    "description:equals:%s" % description
+                ]
+            }
+        )
+        if strict and orders: raise appier.OperationalError(
+            message = "Duplicated order '%s' in Omni" % description
+        )
+
+        # tries to obtain at least one of the customers that match
+        # the email associated with the order in Omni's data source
+        # in case a match then this customer is going to be used as
+        # the owner of the sale otherwise a fallback must be performed
+        customers = api.list_customers(
+            number_records = 1,
+            **{
+                "filters[]" : [
+                    "primary_contact_information.email:equals:%s" % self.email
+                ]
+            }
+        ) if self.email else []
+
+        if customers:
+            # "gathers" the first customer of the sequence as the
+            # existing customer so that it's contents can be explored
+            existing = customers[0]
+
+            # must edit the existing customer information to update
+            # its VAT number with the new one, this way a possible
+            # invoice for the new sale will be "printed" with the
+            # expected new VAT number
+            if self.billing_address.vat_number and\
+                not existing["tax_number"] == self.billing_address.vat_number:
+                api.update_person(
+                    existing["object_id"],
+                    dict(
+                        customer_person = dict(
+                            tax_number = self.billing_address.vat_number
+                        )
+                    )
+                )
+
+            # sets the customer as an existing one by referencing its
+            # object ID and marking parameters as existing
+            customer = dict(
+                object_id = existing["object_id"],
+                _parameters = dict(
+                    type = "existing"
+                )
+            )
+        elif self.account and self.billing_address and self.email:
+            customer = dict(
+                name = self.account.first_name,
+                surname = self.account.last_name,
+                primary_contact_information = dict(
+                    phone_number = self.billing_address.phone_number,
+                    email = self.email
+                ),
+                primary_address = dict(
+                    street_name = self.billing_address.address,
+                    zip_code = self.billing_address.postal_code,
+                    zip_code_name = self.billing_address.city,
+                    country = COUTRIES_MAP.get(self.billing_address.country, None)
+                ),
+                tax_number = self.billing_address.vat_number,
+                observations = "created by Budy",
+                _parameters = dict(
+                    type = "new"
+                )
+            )
+        else:
+            customer = dict(
+                _parameters = dict(
+                    type = "anonymous"
+                )
+            )
+
+        # constructs the complete set of sale lines for the
+        # Omni sale taking into consideration to object ID
+        # of the stored product (assumes Omni imported product)
+        sale_lines = []
+        for line in self.lines:
+            appier.verify(
+                "object_id" in line.product.meta,
+                message = "Product was not imported from Omni, no object ID"
+            )
+            sale_line = dict(
+                merchandise = dict(object_id = line.product.meta["object_id"]),
+                quantity = line.quantity
+            )
+            sale_lines.append(sale_line)
+
+        primary_payment = dict(
+            payment_lines = [
+                dict(
+                    payment_method = dict(_class = "CashPayment"),
+                    amount = dict(value = self.payable)
+                )
+            ]
+        )
+        transaction = dict(
+            description = description,
+            sale_lines = sale_lines,
+            primary_payment = primary_payment
+        )
+        if store_id: transaction["owner"] = dict(object_id = store_id)
+        if self.discount: transaction["discount_vat"] = self.discount
+
+        payload = dict(
+            transaction = transaction,
+            customer = customer
+        )
+
+        try:
+            sale = api.create_sale(payload)
+        except Exception as exception:
+            if hasattr(exception, "error"):
+                self.logger.warn(str(exception.error._data))
+            raise
+
+        # in case an invoice was requested then a proper money sale
+        # slip must be generated for the target sale
+        if invoice:
+            api.issue_money_sale_slip_sale(sale["object_id"])
+
+        # updates the complete set of metadata related with the Omni
+        # import operation so that the order is properly "marked" and
+        # the linking can be properly "explorer"
+        self.meta.update(
+            omni_timestamp = time.time(),
+            omni_sale = sale["object_id"],
+            omni_url = api.base_url + "omni_sam/sales/%s" % sale["object_id"]
+        )
+        self.save()
+
     @appier.link(name = "Export Lines CSV")
     def lines_csv_url(self, absolute = False):
         return appier.get_app().url_for(
@@ -1082,7 +1274,7 @@ class Order(bundle.Bundle):
         if has_store_currency: return self.store.currency_code
         return self.shipping_currency
 
-    def _pay(self, payment_data, strict = True):
+    def _pay(self, payment_data, payment_function = None, strict = True):
         cls = self.__class__
         if self.payable == 0.0: return True
         methods = cls._pmethods()
@@ -1091,12 +1283,15 @@ class Order(bundle.Bundle):
         if not method: raise appier.SecurityError(
             message = "No payment method defined"
         )
-        has_function = hasattr(self, "_pay_" + method)
+        has_function = bool(payment_function) or hasattr(self, "_pay_" + method)
         if not has_function and not strict: return
         if not has_function: raise appier.SecurityError(
             message = "Invalid payment method"
         )
-        function = getattr(self, "_pay_" + method)
+        if self.payment_data and strict: raise appier.SecurityError(
+            message = "Payment data has already been issued"
+        )
+        function = payment_function or getattr(self, "_pay_" + method)
         return function(payment_data)
 
     def _pay_stripe(self, payment_data):
@@ -1328,7 +1523,7 @@ class Order(bundle.Bundle):
         )
         return pay_secret_url
 
-    def _end_pay(self, payment_data, strict = False):
+    def _end_pay(self, payment_data, payment_function = None, strict = False):
         cls = self.__class__
         if self.payable == 0.0: return
         methods = cls._pmethods()
@@ -1338,12 +1533,12 @@ class Order(bundle.Bundle):
         if not method: raise appier.SecurityError(
             message = "No payment method defined"
         )
-        has_function = hasattr(self, "_end_pay_" + method)
+        has_function = payment_function or hasattr(self, "_end_pay_" + method)
         if not has_function and not strict: return
         if not has_function: raise appier.SecurityError(
             message = "Invalid payment method"
         )
-        function = getattr(self, "_end_pay_" + method)
+        function = payment_function or getattr(self, "_end_pay_" + method)
         return function(payment_data)
 
     def _end_pay_stripe(self, payment_data):
@@ -1446,7 +1641,7 @@ class Order(bundle.Bundle):
             )
         return True
 
-    def _cancel(self, cancel_data, strict = False):
+    def _cancel(self, cancel_data, cancel_function = None, strict = False):
         cls = self.__class__
         if self.payable == 0.0: return
         methods = cls._pmethods()
@@ -1454,7 +1649,7 @@ class Order(bundle.Bundle):
         type = cancel_data.get("type", type)
         method = methods.get(type, type)
         if not method: return
-        has_function = hasattr(self, "_cancel_" + method)
+        has_function = cancel_function or hasattr(self, "_cancel_" + method)
         if not has_function and not strict: return
-        function = getattr(self, "_cancel_" + method)
+        function = cancel_function or getattr(self, "_cancel_" + method)
         return function(cancel_data)
